
#include "cache.h"
#include "oj.h"
#include "parser.h"
#include "usual.h"

#define BYTE_OFFSETS_STACK_INC_SIZE 256

VALUE mod = Qnil;

typedef struct _introspect {
    struct _usual usual;
    void (*start)(struct _ojParser *p);
    void (*free)(struct _ojParser *p);
    void (*mark)(struct _ojParser *p);
    void (*open_object)(struct _ojParser *p);
    void (*close_object)(struct _ojParser *p);

    int len;
    int cur;
    long *stack;

} * Introspect;

static void dfree(ojParser p) {
    Introspect d = (Introspect)p->ctx;

    xfree(d->stack);
    // As with the mark() function, this function should still work.
    d->free(p);
}

static void start(ojParser p) {
    Introspect d = (Introspect)p->ctx;

    d->start(p);

    d->stack = ALLOC_N(long, BYTE_OFFSETS_STACK_INC_SIZE);
    d->len = BYTE_OFFSETS_STACK_INC_SIZE;
    d->cur = 0;
}

static void open_object(ojParser p) {
    Introspect d = (Introspect)p->ctx;

    d->open_object(p);

    // TBD add offset
}

static void close_object(ojParser p) {
    Introspect d = (Introspect)p->ctx;

    d->close_object(p);

    // TBD set offset on hash
}

static VALUE new_parser(VALUE self, VALUE ropts) {
    volatile VALUE pv = oj_parser_new();
    ojParser       p  = (ojParser)DATA_PTR(pv);
    Introspect     d  = ALLOC(struct _introspect);

    oj_init_usual(p, &d->usual);

    // Set the options before switching to the Introspect delegate.
    p->ctx    = (void *)d;

    d->free = p->free;
    p->free   = dfree;

    d->start = p->start;
    p->start  = start;

    // We are cheating with the mark, free, and options functions. Since
    // struct _usual is the first member of struct _introspect the cast to
    // Usual in the usual.c mark() function should still work just fine.

    d->open_object = p->funcs[TOP_FUN].open_object;
    p->funcs[TOP_FUN].open_object = open_object;

    d->close_object = p->funcs[TOP_FUN].close_object;
    p->funcs[TOP_FUN].close_object = close_object;

    Check_Type(ropts, T_HASH);
    oj_parser_set_option(p, ropts);

    return pv;
}

void Init_introspect(void) {
    mod = rb_define_module("Introspect");
    rb_gc_register_address(&mod);

    rb_define_module_function(mod, "parser", new_parser, 1);
}
